/*-----------------------------------------------
p_surf.inc
An include file for generating parameterized
"uv" surfaces with POV-Ray 3.1.
                    

The following requires that a vector valued macro
named 'XYZ' and variables 
        uusteps,        (uu mesh resolution)
	vvsteps,        (vv mesh resolution) 
        uumin,          (low end of uu)  
        uumax,          (high end of uu)
        vvmin,          (low end of vv)   
        vvmax           (high end of vv)  
have been previously defined.
                                            
                                            
If an additional macro 'mytexture1(uu,vv)'
been defined, it can be used to texture individual
triangles.
	
A particular caveat for making smooth triangles
is that even when uumin, uumax,vvmin, and vvmax have
been carefully set to avoid trouble (square roots of
negative numbers, etc), in fact the routine for calculating 
normal vectors explores areas of the u-v plane
outside of the uumin->uumax X vvmin->vvmax rectangle.
This can lead to errors if XYZ(uu,vv) happens to explore a 
'sore spot'. A way to avoid these problems is to have 
XYZ(uu,vv) return a 'good' value if <uu,vv> are in a bad place.

Also included is the #macro 
        p_edge(whichEdge,edgeDivisions,edgeRadius) 
which puts cylinders along a border of the surface. whichEdge
is either MINUU, MAXUU, MINVV, or MAXVV. This tells
which border to put the edge on. edgeDivisions is how
many cylinders to use, and edgeRadius is the radius of
the cylinders (the width of the border)
-----------------------------------------------*/
#if (version<3.1)
	#error "This file requires features of POV-Ray version 3.1 (or later?)"
#end

#ifdef(WriteRAWto)
    #fopen RAWFileHandle WriteRAWto write
#end
#if(MakeTriangles) //MakeTriangles set to 1 by the calling file
	#ifdef(WriteTrianglesTo)	// WriteTrianglesTo is name for 
						// triangle data file to write to
		#fopen DataFileHandle WriteTrianglesTo write
		#if ( !defined(DataFileHandle))
			#warning concat("\nUnable to open ",WriteTrianglesTo)
			#undef WriteTrianglesTo
		#end
	#end
     
	// To speed up triangle and normals calculations,
	// we will compute the surface points and store
	// them in XYZ_Array. This is larger than just
	// uusteps * vvsteps because we need surface points
	// beyond the declared edges in order to calculate
	// normals.
	#local XYZ_Array=array[uusteps+3][vvsteps+3]
     
	// If we want smooth triangles, we will pre-compute
	// and store the normals, too.
	#if(use_smooth_triangles)
		#local Normal_Array=array[uusteps+3][vvsteps+3]
	#end

	// uu,vv lattice interval sizes
	#local duu= (uumax-uumin)/uusteps;//u-increment per step
	#local dvv= (vvmax-vvmin)/vvsteps;//v-increment per step

	#debug "Calculating surface points"
	#local vv= vvmin-dvv;  
	#local vvcount = 0;
	#while (vvcount<vvsteps+3) // vv loop
		#local uucount = 0;
		#local uu = uumin-duu;
		#while (uucount<uusteps+3) // uu loop
			#local XYZ_Array[uucount][vvcount]=XYZ(uu,vv);
			#local uucount=uucount+1;
			#local uu = uu+duu;
		#end  //end uu loop  		

		#local vvcount = vvcount + 1;
		#local vv = vv+dvv;
	#end	//end vv loop
     
	#if(use_smooth_triangles)
		#debug "\nCalculating normals"
		#local vv= vvmin;  
		#local vvcount = 1;

		#while (vvcount<=vvsteps+1) // vv loop
			#local uucount=1;
			#local uu = uumin;
			#while (uucount<=uusteps+1) // uu loop			
				#local p1=XYZ_Array[uucount][vvcount];
				#local p2=XYZ_Array[uucount+1][vvcount];
				#local p4=XYZ_Array[uucount][vvcount+1];
				#local p1a=XYZ_Array[uucount-1][vvcount];
				#local p1b=XYZ_Array[uucount][vvcount-1];
/*  
                  p4
                  | 
                  | 
                  | 
         p1a-----p1------p2
                  |
                  |         
     /|\          |       
      v          p1b     
       u->    
*/
				// Calculate normal at p1:
				#local n1=vcross(p1-p1b,p1-p1a);
				#local n2=vcross(p1-p2,p1-p1b);
				#local n3=vcross(p1-p4,p1-p2);
				#local n4=vcross(p1-p1a,p1-p4);
				#local Normal_Array[uucount][vvcount]=n1+n2+n3+n4;
     
				#local uucount=uucount+1;
				#local uu = uu+duu;
			#end  //end uu loop  		
     		
			#local vvcount = vvcount + 1;
			#local vv = vv+dvv;
		#end	//end vv loop
	#end //end if(use_smooth_triangles)
          
	#debug "\nMaking "
	#if (defined(DataFileHandle))
		#debug "and writing "
	#end
	#debug "triangles"

	#declare NumberOfTriangles=0;
	#local vvcount = 1;
	#declare p_surface =
	mesh{
	#while (vvcount<=vvsteps) // vv loop
		//#debug concat(str(vvcount,0,0)," / ",str(vvsteps,0,0),"\r")
                             
		#local uucount=1;
     
		#while (uucount<=uusteps) // uu loop			
                             
			// get triangle vertices p1,p2,p3,p4
			// Each uuXvv rectangle gives 2 triangles (tr1 and tr2)

/*
                 p4a     p3b
                  .       .
                  .       .
         p4b.... p4------p3.... p3a
                  | tr2.  |
                  |  .    |
                  |.  tr1 |
         p1a.... p1------p2.... p2b
                  .       .  
     /|\          .       .
      v          p1b     p2a   
       u->
         
*/                             
			#local p1=XYZ_Array[uucount][vvcount];
			#local p2=XYZ_Array[uucount+1][vvcount];
			#local p3=XYZ_Array[uucount+1][vvcount+1];
			#local p4=XYZ_Array[uucount][vvcount+1];
			
			// if a texture macro named mytexture1
			// has been declared, use it for
			// the triangles
			#ifdef(mytexture1)
				#local T1=mytexture1(uucount,vvcount)
			#else
			        #ifdef(T1)
			                #undef T1
			        #end
			#end
			
                        
			#if(use_smooth_triangles)
				// get normals:
				#local n1=Normal_Array[uucount][vvcount];
				#local n2=Normal_Array[uucount+1][vvcount];
				#local n3=Normal_Array[uucount+1][vvcount+1];
				#local n4=Normal_Array[uucount][vvcount+1];
				//make smooth triangles
                                
				smooth_triangle  {
					p1, n1,
					p2, n2,
					p3, n3
					#ifdef(T1)
						texture{T1}
					#end
				}
     
				smooth_triangle  {
					p1, n1,
					p3, n3,
					p4, n4
					#ifdef(T1)
						texture{T1}
					#end
				}
				#if(defined(DataFileHandle))
     					
					#write (DataFileHandle,p1,",",n1,",",p2,",",n2,",",p3,",",n3,",","\n")
					#write (DataFileHandle,p1,",",n1,",",p3,",",n3,",",p4,",",n4,",","\n")
     
				#end 
				#ifdef(RAWFileHandle) 
                                  #declare rawstring=concat (str(p1.x,0,3)," ",str(p1.y,0,3)," ",str(p1.z,0,3)," ", str(p2.x,0,3)," ",str(p2.y,0,3)," ",str(p2.z,0,3)," ",str(p3.x,0,3)," ",str(p3.y,0,3)," ",str(p3.z,0,3)," ")
                                  #write (RAWFileHandle,rawstring) 
                                  //#declare rawstring=concat (str(n1.x,0,3)," ",str(n1.y,0,3)," ",str(n1.z,0,3)," ", str(n2.x,0,3)," ",str(n2.y,0,3)," ",str(n2.z,0,3)," ",str(n3.x,0,3)," ",str(n3.y,0,3)," ",str(n3.z,0,3),"\n")
                                  //#write (RAWFileHandle,rawstring) 
                                  #write (RAWFileHandle,"\n")
                                  
                                  #declare rawstring=concat (str(p1.x,0,3)," ",str(p1.y,0,3)," ",str(p1.z,0,3)," ", str(p3.x,0,3)," ",str(p3.y,0,3)," ",str(p3.z,0,3)," ",str(p4.x,0,3)," ",str(p4.y,0,3)," ",str(p4.z,0,3)," ")
                                  #write (RAWFileHandle,rawstring) 
                                  //#declare rawstring=concat (str(n1.x,0,3)," ",str(n1.y,0,3)," ",str(n1.z,0,3)," ", str(n3.x,0,3)," ",str(n3.y,0,3)," ",str(n3.z,0,3)," ",str(n4.x,0,3)," ",str(n4.y,0,3)," ",str(n4.z,0,3),"\n")
                                  //#write (RAWFileHandle,rawstring)
                                  #write (RAWFileHandle,"\n")
                                #end
			#else // 'flat' triangles
     				
				triangle  {
					p1,p2,p3
					#ifdef(mytexture1)
						texture{T1}
					#end
				}
				triangle  {
					p1,p3,p4
					#ifdef(mytexture1)
						texture{T1}
					#end
				}  
				#if(defined(DataFileHandle))
     					
					#write (DataFileHandle,p1,",",p2,",",p3,",","\n")
					#write (DataFileHandle,p1,",",p3,",",p4,",","\n")
     
				#end 
				#ifdef(RAWFileHandle) 
                                  #declare rawstring=concat (str(p1.x,0,3)," ",str(p1.y,0,3)," ",str(p1.z,0,3)," ", str(p2.x,0,3)," ",str(p2.y,0,3)," ",str(p2.z,0,3)," ",str(p3.x,0,3)," ",str(p3.y,0,3)," ",str(p3.z,0,3),"\n")
                                  #write (RAWFileHandle,rawstring)
                                  #declare rawstring=concat (str(p1.x,0,3)," ",str(p1.y,0,3)," ",str(p1.z,0,3)," ", str(p3.x,0,3)," ",str(p3.y,0,3)," ",str(p3.z,0,3)," ",str(p4.x,0,3)," ",str(p4.y,0,3)," ",str(p4.z,0,3),"\n")
                                  #write (RAWFileHandle,rawstring)
                                #end            
			#end
			#declare NumberOfTriangles=NumberOfTriangles+2;
			#local uucount=uucount+1;
     		
		#end  //end uu loop  		
     		
		#local vvcount = vvcount + 1;
     		
	#end	//end vv loop
	} // end of mesh
	#ifdef(WriteTrianglesTo)
		#fclose DataFileHandle
		#debug concat("\nWrote ",str(NumberOfTriangles,0,0)," triangles to \"",WriteTrianglesTo,"\"\n")
	#end

#else // if not making triangles, we read them

	#fopen DataFileHandle ReadTrianglesFrom read
	#debug concat("Reading triangles from \"",ReadTrianglesFrom,"\" ... ")
        #declare uucount=0;
        #declare vvcount=0;
        #declare tgridcount=0;
	#declare p_surface=
	mesh{
		#declare NumberOfTriangles=0;
		#while(defined(DataFileHandle))
		        #ifdef(mytexture1)
				#local T1=mytexture1(uucount,vvcount)
			#end
			#if(use_smooth_triangles)
				#read(DataFileHandle,p1,n1,p2,n2,p3,n3)
				smooth_triangle{
					p1,n1,p2,n2,p3,n3  
					#ifdef(T1)
						texture{T1}
					#end
				} 
				#ifdef(RAWFileHandle) 
                                  #declare rawstring=concat (str(p1.x,0,3)," ",str(p1.y,0,3)," ",str(p1.z,0,3)," ", str(p2.x,0,3)," ",str(p2.y,0,3)," ",str(p2.z,0,3)," ",str(p3.x,0,3)," ",str(p3.y,0,3)," ",str(p3.z,0,3)," ")
                                  #write (RAWFileHandle,rawstring) 
                                  //#declare rawstring=concat (str(n1.x,0,3)," ",str(n1.y,0,3)," ",str(n1.z,0,3)," ", str(n2.x,0,3)," ",str(n2.y,0,3)," ",str(n2.z,0,3)," ",str(n3.x,0,3)," ",str(n3.y,0,3)," ",str(n3.z,0,3),"\n")
                                  //#write (RAWFileHandle,rawstring)
                                  #write (RAWFileHandle,"\n")
                                  
                                  
                                #end
				
			#else
				#read(DataFileHandle,p1,p2,p3)
				triangle{
					p1,p2,p3 
					#ifdef(T1)
						texture{T1}
					#end
				} 
				
                                 #ifdef(RAWFileHandle)
                            
                                    #declare rawstring=concat (str(p1.x,0,3)," ",str(p1.y,0,3)," ",str(p1.z,0,3)," ", str(p2.x,0,3)," ",str(p2.y,0,3)," ",str(p2.z,0,3)," ",str(p3.x,0,3)," ",str(p3.y,0,3)," ",str(p3.z,0,3),"\n ")
                                    #write (RAWFileHandle,rawstring)
                                 #end
			#end              
			
			#declare NumberOfTriangles=NumberOfTriangles+1;
			#declare tgridcount=tgridcount+1;
			
			#if(tgridcount=2)
			        #declare uucount=uucount+1;
			        #if(uucount=uusteps)
			            #declare vvcount=vvcount+1;
			            #declare uucount=0;
			        #end     
			        #declare tgridcount=0;
			#end
			
		#end
		#debug concat("(",str(NumberOfTriangles-1,0,0)," read)\n")
	}


#end


// p_edge is a utility macro that puts a border
// made of little cylinders along an 'edge' of the
// surface (along a minimum or maximum of uu or vv
#declare MINUU=1;
#declare MAXUU=2;
#declare MINVV=4;
#declare MAXVV=8;
#macro p_edge( whichEdge,edgeDivisions,edgeRadius)
  #switch (whichEdge)
          #case (MINUU)
                #local edgemin=vvmin;
                #local edgemax=vvmax;
                #local edgeu=uumin;
                #local edgev=vvmin;
          #break
          #case (MAXUU)
                #local edgemin=vvmin;
                #local edgemax=vvmax;
                #local edgeu=uumax;
                #local edgev=vvmin;
          #break 
        
          #case (MINVV)
                #local edgemin=uumin;
                #local edgemax=uumax;
                #local edgeu=uumin;
                #local edgev=vvmin;
          #break 
          #case (MAXVV)
                #local edgemin=uumin;
                #local edgemax=uumax;
                #local edgeu=uumin;
                #local edgev=vvmax;
          #break 
  #end
  #local uu=edgeu;
  #local vv=edgev;
  #local de=(edgemax-edgemin)/edgeDivisions;
  #local edgeN=0;
  union{
        #local lastvertex=XYZ(uu,vv);
        #while(edgeN<=edgeDivisions)
               #local thisvertex=XYZ(uu,vv);
               #if(lastvertex.x!=thisvertex.x | lastvertex.y!=thisvertex.y | lastvertex.z!=thisvertex.z)
                         cylinder{lastvertex,thisvertex,edgeRadius }
               #end
               sphere{thisvertex,edgeRadius}
               #local lastvertex=thisvertex;
               #switch(whichEdge)
                   #case (MINUU)
                        #local vv=vv+de;
                   #break
                   #case (MAXUU)
                        #local vv=vv+de;
                   #break
                   #case (MINVV)
                        #local uu=uu+de;
                   #break
                   #case (MAXUU)
                        #local uu=uu+de;
                   #break
               #end
               #local edgeN=edgeN+1;
        #end
  }
#end

  
#ifdef(RAWFileHandle)
    #fclose RAWFileHandle

#end  